// ==============================================================================//
// Телеметр (персональная метеостанция) на ESP8266 NodeMcu V3                    //
// ==============================================================================//
// Используется 4 сенсора:                                                       // 
// 1. BME280  - улица ( канал: https://thingspeak.com/channels/654121 )          //
// 2. DHT22   - гараж                                                            //
// 3. DHT22   - погреб 1                                                         //
// 4. DHT22   - погреб 2 (по этому датчику включается подогрев)                  //
// ==============================================================================//
// Данные передаются сразу на 2 сервера: MQTT cloudmqtt.com и thingspeak.com     //
// 2018-12-20: Добавлена поддержка IoT Manager                                   //
// ==============================================================================//
// Автор: Александр Разживин, kotyara12@yandex.ru, http://ravsoft2004.narod.ru   //
// ==============================================================================//

#include <EEPROM.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h> 
#include <PubSubClient.h>
#include "Wire.h"
#include "Adafruit_BME280.h"
#include <Adafruit_Sensor.h>
#include "DHTesp.h"

// =============================================================================================================================
// == Настройки для конкретного исполнения =====================================================================================
// =============================================================================================================================

// Выводы, к которым подкючена периферия
// #define pinSCL D1
// #define pinSCA D2
#define pinLED    D3
#define pinLEDINT D4
#define pinSENS2  D5
#define pinSENS3  D6
#define pinSENS4  D7
#define pinRELAY  D8

// Параметры сетей WiFi
const char*    wifi1_ssid           = "*********";        // Имя WiFi сети 1
const char*    wifi1_pass           = "*********";        // Пароль WiFi сети 1
const char*    wifi2_ssid           = "*********";        // Имя WiFi сети 2
const char*    wifi2_pass           = "*********";        // Пароль WiFi сети 2
const char*    wifi3_ssid           = "*********";        // Имя WiFi сети 3
const char*    wifi3_pass           = "*********";        // Пароль WiFi сети 4

// Параметры сервера MQTT
const char*    mqttServer          = "m88.cloudmqtt.com"; // Имя сервера MQTT
const int      mqttPort            = *****;               // Порт для подключения к серверу MQTT
const char*    mqttUser            = "user";              // Логин от сервера MQTT
const char*    mqttPass            = "pass";              // Пароль от сервера MQTT
const char*    mqttPrefix          = "village";           // Префикс всех топиков для сервера MQTT
const char*    mqttDeviceId        = "telemeter";         // Идентификатор клиента MQTT

// Топики универсальные
const char*    mqttDeviceState     = "status";            // Публикация состояния устройства и WillMessage
const char*    mqttSettings        = "settings";          // Параметры устройства
const char*    mqttSyncMode        = "sync_mode";         // Включение или отключение режима синхронизации
const char*    mqttSyncRequest     = "sync";              // Запрос статуса устройства

// Топики для IoT Manager
const char*    mqttIoT             = "iotmanager";        // Префикс для IoT Manager
const char*    mqttPushIds         = "ids";               // Получение идектификаторов Push-уведомлений
const char*    mqttDevice          = "device";            // Получение информации о мобильных устройствах
const char*    mqttResuest         = "request";           // Получение запросов на отправку конфига
const char*    mqttControl         = "control";           // Получение управляющих команд виджетов
const char*    mqttStatus          = "status";            // Публикация состояния виджета
 
// Топики сенсоров и команд
const char*    mqttSensor1         = "outdoor";           // Улица
const char*    mqttSensor2         = "garage";            // Гараж 
const char*    mqttSensor3         = "cellar_up";         // Погреб 1
const char*    mqttSensor4         = "cellar_down";       // Погреб 2

const char*    mqttHeater          = "heater";            // Нагреватель
const char*    mqttHeaterState     = "status";            // Состояние нагревателя
const char*    mqttHeaterSensor    = mqttSensor4;         // Сенсор, который управляет нагревателем

const char*    mqttCommands        = "commands";          // Управлящие команды
const char*    mqttPubInterval     = "interval_publ";     // Интервал публикации значений сенсоров
const char*    mqttThingSpeak      = "thingspeak";        // Режим публикации на ThingSpeak
const char*    mqttHeaterMode      = "mode_int";          // Режим управления нагревателем
const char*    mqttHeaterMin       = "temp_min";          // Минимальная температура на автоматического режима
const char*    mqttHeaterMax       = "temp_max";          // Максимальная температура на автоматического режима

// Точики значений сенсоров
const char*    mqttSensorState     = "sensor_state";      // Статус сенсора
const char*    mqttSensorType      = "sensor_type";       // Статус сенсора
const char*    mqttPressurePa      = "pressure_pascal";   // Давление в паскалях
const char*    mqttPressureHg      = "pressure_mmhg";     // Давление в мм рт ст
const char*    mqttTemperature     = "temp";              // Температура
const char*    mqttHumidity        = "humidity";          // Влажность
const char*    mqttHeatIndex       = "heat_index";        // Индекс
const char*    mqttDewPoint        = "dew_point";         // Точка росы
const char*    mqttRSSI            = "rssi";              // Уровень сигнала WiFi

// Команды
const char*    mqttCmdReset        = "reset";             // Перезапуск устройства

// Параметры ThingSpeak
const char*    ts_ApiServer        = "api.thingspeak.com";
const char*    ts_ApiPost          = "GET /update?%s HTTP/1.1\r\n";
const char*    ts_ApiHost          = "Host: %s\r\n";

// Каналы ThingSpeak
const char*    ts_APIKey_Sensor1   = "****************";   // Улица
const char*    ts_APIKey_Sensor2   = "****************";   // Гараж
const char*    ts_APIKey_Sensor3   = "****************";   // Погреб 1
const char*    ts_APIKey_Sensor4   = "****************";   // Погреб 2

// =============================================================================================================================
// =============================================================================================================================
// =============================================================================================================================

// Прочие константы 
const char*    stateOK             = "OK";
const char*    stateDisconnected   = "DISCONNECTED";

// Рабочие параметры
unsigned long  lastPublish         = 0;                   // Время последней публикации данных
unsigned long  lastPublishTS       = 0;                   // Время последней публикации данных на сервере ThingSpeak
uint16_t       intvPublish         = 300;                 // Интервал сбора данных, секунд
bool           wifiState           = false;               // Текущее состояние WiFi-подключения
bool           ledState            = false;               // Статус LED для мигания светодиодиком
bool           syncRequests        = false;               // Статус запросов на синхронизацию
int8_t         heater_Mode         = 0;                   // Режим управления нагревателем
bool           heater_State        = false;               // Текущий режим нагревателя
int8_t         heater_AutoOn       = 0;                   // Температура, при которой включается нагреватель
int8_t         heater_AutoOff      = 5;                   // Температура, при которой выключается нагреватель
bool           stateSensor1        = false;               // Статус наружнего сенсора
bool           publThingSpeak      = true;                // Публикация данных на сервере ThingSpeak
String         mqttWill            = "";                  // Топик Will Message

// Адреса параметров
const int16_t  heater_Mode_Addr    = 0;                   // Режим управления нагревателем
const int16_t  heater_AutoOn_Addr  = 1;                   // Температура, при которой включается нагреватель
const int16_t  heater_AutoOff_Addr = 2;                   // Температура, при которой выключается нагреватель
const int16_t  intvPublish_Addr    = 3;                   // Интервал сбора данных
const int16_t  publThingSpeak_Addr = 5;                   // Публикация данных на сервере ThingSpeak
const int16_t  syncRequests_Addr   = 6;                   // Статус запросов на синхронизацию

// Прочие параметры 
const bool     sensorsRetained     = true;                // Retained для сенсоров
const byte     sensorsQoS          = 0;                   // QoS для сенсоров
const bool     deviceRetained      = true;                // Retained для статуса устройства
const byte     deviceQoS           = 2;                   // QoS для статуса устройства
const bool     heaterRetained      = true;                // Retained для нагревателя
const byte     heaterQoS           = 2;                   // QoS для нагревателя
const long     waitPublMQTT        = 500;                 // Задержка при публикации только в MQTT
const long     waitPublThingSpeak  = 20000;               // Задержка при публикации на сервере ThingSpeak

// Переменные ThingSpeak
const char*    ts_ApiKey           = "api_key=";
const char*    ts_ApiField         = "&field";

ESP8266WiFiMulti wifiMulti;
WiFiClient wifiMqtt;
PubSubClient mqttClient(wifiMqtt);

Adafruit_BME280 bmeSensor1;
DHTesp dhtSensor2;
DHTesp dhtSensor3;
DHTesp dhtSensor4;

// =============================================================================================================================
// == Чтение и запись в EEPROM =================================================================================================
// =============================================================================================================================
void EEPROM_bool_write(int addr, bool value) {
  byte raw = 0;
  if (value) { raw = 1; };
  EEPROM.write(addr, raw);
}

bool EEPROM_bool_read(int addr, bool def_value) {
  byte raw = EEPROM.read(addr);
  if (raw == -1) {
    EEPROM_bool_write(addr, def_value);
    return def_value;
  }
  return raw > 0;
}

void EEPROM_uint8_write(int addr, uint8_t value) {
  EEPROM.write(addr, value);
}

uint8_t EEPROM_uint8_read(int addr, uint8_t def_value) {
  uint8_t raw = EEPROM.read(addr);
  if (raw == -1) {
    EEPROM_uint8_write(addr, def_value);
    return def_value;
  }
  return raw;
}

void EEPROM_uint16_write(int addr, uint16_t value) {
  byte raw[2];
  (uint16_t&)raw = value;
  for(byte i = 0; i < 2; i++) EEPROM.write(addr+i, raw[i]);
}

uint16_t EEPROM_uint16_read(int addr, uint16_t def_value) {   
  byte raw[2];
  for(byte i = 0; i < 2; i++) raw[i] = EEPROM.read(addr+i);
  if (raw[0] = -1) {
    EEPROM_uint16_write(addr, def_value);
    return def_value;
  }
  uint16_t &value = (uint16_t&)raw;
  return value;
}

// =============================================================================================================================
// == MQTT: сервисные функции ==================================================================================================
// =============================================================================================================================

void mqttWait(const unsigned long delayWait) {
  unsigned long waitStart = millis();

  while ((millis() - waitStart) < delayWait) {
    if (mqttClient.connected()) {
      mqttClient.loop();
    };
    
    delay(100);
    yield();
  };
}

String mqttGetDeviceTopic(const char* topic) {
  return String(mqttPrefix) + "/" + mqttDeviceId + "/" + topic;
}

String mqttGetDeviceTopic(const char* topic1, const char* topic2) {
  return String(mqttPrefix) + "/" + mqttDeviceId + "/" + topic1 + "/" + topic2;
}

// =============================================================================================================================
// == MQTT: публикация значений ================================================================================================
// =============================================================================================================================
void mqttPublish(const char* topic, const char* payload, const boolean retained, const byte qos) {
  String topicPub = mqttGetDeviceTopic(topic);
  
  Serial.print("MQTT :: publish value to topic '");
  Serial.print(topicPub);
  Serial.print("': ");
  Serial.println(payload);

  if (mqttClient.connected()) {
    MQTT::Publish mqttPub = MQTT::Publish(topicPub.c_str(), payload);
    mqttPub.set_retain(retained);
    mqttPub.set_qos(qos);

    if (!mqttClient.publish(mqttPub)) {
      mqttWait(waitPublMQTT);
      if (!mqttClient.publish(mqttPub)) {
        Serial.print("ERROR :: publish error!");
      };
    };
  }
  else {
    Serial.print("ERROR :: connection lost!");
  };
  
  mqttWait(waitPublMQTT);
}

void mqttPublishFloat(const char* topic, const char* subtopic, const float value, const boolean retained, const byte qos) {
  String topicFull = String(topic);
  if (subtopic != "") topicFull = topicFull + "/" + subtopic;

  mqttPublish(topicFull.c_str(), String(value).c_str(), retained, qos);
};

void mqttPublishChar(const char* topic, const char* subtopic, const char* value, const boolean retained, const byte qos) {
  String topicFull = String(topic);
  if (subtopic != "") topicFull = topicFull + "/" + subtopic;
  
  mqttPublish(topicFull.c_str(), value, retained, qos);
};

void mqttPublishBool(const char* topic, const char* subtopic, const bool value, const boolean retained, const byte qos) {
  if (value) {
    mqttPublishChar(topic, subtopic, "1", retained, qos);
  }
  else {
    mqttPublishChar(topic, subtopic, "0", retained, qos);
  };
}

// =============================================================================================================================
// == ThingSpeak: публикация значений ========================================================================================== 
// =============================================================================================================================
void tsPublishData(const char* request) {
  // Для отправки запроса на ThingSpeak используем локальный экземпляр WiFiClient
  WiFiClient tsClient;
  
  // Ждем, пока пройдет 20 секунд с момента последнего запроса
  if ((millis() - lastPublishTS) < waitPublThingSpeak) {
    Serial.print("ThingSpeak :: wait ");
    Serial.print(waitPublThingSpeak - (millis() - lastPublishTS));
    Serial.println(" ms...");

    while ((millis() - lastPublishTS) < waitPublThingSpeak) {
      if (mqttClient.connected()) {
        mqttClient.loop();
      };
      
      delay(100);
      yield();
    };
  }
  
  Serial.print("ThingSpeak :: send data '");
  Serial.print(request);
  Serial.println("'");

  // Отправляем HTTP-запрос
  if (tsClient.connect(ts_ApiServer, 80)) {
    tsClient.printf(ts_ApiPost, request);
    tsClient.printf(ts_ApiHost, ts_ApiServer);
    tsClient.println("User-Agent: ESP8266 (nothans)/1.0");
    tsClient.println("Connection: close");
    tsClient.println();

    // Читаем результат (первую строку отправленного заголовка)
    Serial.print("ThingSpeak :: response code: ");
    Serial.println(tsClient.readStringUntil('\n'));

    tsClient.stop();
  }
  else {
    Serial.print("#ERROR# Failed connection to ");
    Serial.println(ts_ApiServer);
  }

  // Запоминаем время последнего соединения
  lastPublishTS = millis();
}

// =============================================================================================================================
// == Публикация различных статусов ============================================================================================
// =============================================================================================================================
void mqttPublishDeviceState() {
  mqttPublishChar(mqttDeviceState, "", "1", deviceRetained, deviceQoS);  
}

void mqttPublishHeaterState() {
  mqttPublishBool(mqttHeater, mqttHeaterState, heater_State, heaterRetained, heaterQoS);  
}

// =============================================================================================================================
// == Публикация данных сенсора BME280 =========================================================================================
// =============================================================================================================================
void mqttPublishSendorBME(Adafruit_BME280 bme, const char* sensor_name, const bool sensor_state, const char* keyAPI, const bool pub_rssi) {
  if (sensor_state) {
    mqttPublishChar(sensor_name, mqttSensorType, "BME280", sensorsRetained, sensorsQoS);
    mqttPublishChar(sensor_name, mqttSensorState, stateOK, sensorsRetained, sensorsQoS);
    
    // Считываем параметры из датчика
    float pressure = bme.readPressure();
    float humidity = bme.readHumidity();
    float temperature = bme.readTemperature();
    float heatindex = dhtSensor2.computeHeatIndex(temperature, humidity, false);
    float dewpoint = dhtSensor2.computeDewPoint(temperature, humidity, false);

    // Публикуем полученные данные на MQTT брокере
    mqttPublishFloat(sensor_name, mqttPressurePa, pressure, sensorsRetained, sensorsQoS);
    mqttPublishFloat(sensor_name, mqttPressureHg, pressure / 133.321995, sensorsRetained, sensorsQoS);
    mqttPublishFloat(sensor_name, mqttTemperature, temperature, sensorsRetained, sensorsQoS);
    mqttPublishFloat(sensor_name, mqttHumidity, humidity, sensorsRetained, sensorsQoS);
    mqttPublishFloat(sensor_name, mqttHeatIndex,heatindex, sensorsRetained, sensorsQoS);
    mqttPublishFloat(sensor_name, mqttDewPoint, dewpoint, sensorsRetained, sensorsQoS);

    // Добавляем уровень сигнала WiFi сети
    if (pub_rssi) mqttPublishFloat(sensor_name, mqttRSSI, WiFi.RSSI(), sensorsRetained, sensorsQoS);

    // Публикуем полученные данные на ThingSpeak
    if (publThingSpeak && (keyAPI != "")) {
      // Формируем GET запрос
      String payload = String(ts_ApiKey) + keyAPI;
             payload = payload + ts_ApiField + "1=" + String(pressure);
             payload = payload + ts_ApiField + "2=" + String(pressure / 133.321995);
             payload = payload + ts_ApiField + "3=" + String(temperature);
             payload = payload + ts_ApiField + "4=" + String(humidity);
             payload = payload + ts_ApiField + "5=" + String(heatindex);
             payload = payload + ts_ApiField + "6=" + String(dewpoint);
             if (pub_rssi) payload = payload + ts_ApiField + "7=" + String(WiFi.RSSI()); 
      // Отправляем запрос на сервер
      tsPublishData(payload.c_str());
    }
  }
  else {
    mqttPublishChar(sensor_name, mqttSensorType, "---", sensorsRetained, sensorsQoS);
    mqttPublishChar(sensor_name, mqttSensorState, stateDisconnected, sensorsRetained, sensorsQoS);
  }

  Serial.println();
}

// == Публикация данных сенсора DHT11/DHT22 ======================================================================================
void mqttPublishSendorDHT(DHTesp dht, const char* sensor_name, const char* keyAPI, const bool pub_rssi, const bool heater) {
  // Считываем температуру и влажность из датчика
  TempAndHumidity newValues = dht.getTempAndHumidity();
  
  // Если датчик выдал ошибку ERROR_TIMEOUT, ждем и пробуем еще раз
  if (dht.getStatus() == DHTesp::ERROR_TIMEOUT) { 
    delay(dht.getMinimumSamplingPeriod()); 
    newValues = dht.getTempAndHumidity();
  }

  // Публикуем тип и статус датчика
  const char* sensor_type = "---";
  if (dht.getStatus() != DHTesp::ERROR_TIMEOUT) { 
    switch (dht.getModel()) {
      case DHTesp::DHT11: 
        sensor_type = "DHT11";
        break;
      case DHTesp::DHT22: 
        sensor_type = "DHT22";
        break;
      case DHTesp::AM2302: 
        sensor_type = "AM2302";
        break;
      case DHTesp::RHT03: 
        sensor_type = "RHT03";
        break;
      defailt:  
        sensor_type = "???";
    };
  };
  mqttPublishChar(sensor_name, mqttSensorType, sensor_type, sensorsRetained, sensorsQoS);
  mqttPublishChar(sensor_name, mqttSensorState, dht.getStatusString(), sensorsRetained, sensorsQoS);
  
  // Если удалось получить данные, публикуем их
  if (dht.getStatus() == DHTesp::ERROR_NONE) {
    float heatindex = dht.computeHeatIndex(newValues.temperature, newValues.humidity, false);
    float dewpoint = dht.computeDewPoint(newValues.temperature, newValues.humidity, false);
    
    // Отправляем основные данные
    mqttPublishFloat(sensor_name, mqttTemperature, newValues.temperature, sensorsRetained, sensorsQoS);
    mqttPublishFloat(sensor_name, mqttHumidity, newValues.humidity, sensorsRetained, sensorsQoS);
    mqttPublishFloat(sensor_name, mqttHeatIndex, heatindex, sensorsRetained, sensorsQoS);
    mqttPublishFloat(sensor_name, mqttDewPoint, dewpoint, sensorsRetained, sensorsQoS);
    
    // Добавляем уровень сигнала WiFi сети
    if (pub_rssi) { mqttPublishFloat(sensor_name, mqttRSSI, WiFi.RSSI(), sensorsRetained, sensorsQoS); };
    
    // Если нужно управлять нагревателем, проверяем границы переключения
    if ((heater) && (heater_Mode == 2)) {
      if ((heater_State == false) && (newValues.temperature <= heater_AutoOn)) {
        Serial.println();
        heaterToggle(true, true);
      };
      if ((heater_State == true) && (newValues.temperature >= heater_AutoOff)) {
        Serial.println();
        heaterToggle(false, true);
      };
    };

    // Публикуем полученные данные на ThingSpeak
    if (publThingSpeak && (keyAPI != "")) {
      // Формируем GET запрос
      String payload = String(ts_ApiKey) + keyAPI;
             payload = payload + ts_ApiField + "1=" + String(newValues.temperature);
             payload = payload + ts_ApiField + "2=" + String(newValues.humidity);
             payload = payload + ts_ApiField + "3=" + String(heatindex);
             payload = payload + ts_ApiField + "4=" + String(dewpoint);

      // Добавляем уровень сигнала WiFi сети
      if (pub_rssi) { 
        payload = payload + ts_ApiField + "5=" + String(WiFi.RSSI()); 
      };
      
      // Добавляем статус нагревателя
      if (heater) {
        if (heater_State) {
          payload = payload + ts_ApiField + "5=1";
        }
        else {
          payload = payload + ts_ApiField + "5=0";     
        };
      };

      // Отправляем запрос на сервер
      tsPublishData(payload.c_str());
    };
  };
  
  Serial.println();
}

// =============================================================================================================================
// == Считываение измерений с датчиков и публикация результатов измерений ======================================================
// =============================================================================================================================
void mqttPublishSensorsData() {
  Serial.println("Data publishing started...");
  Serial.println();

  mqttPublishHeaterState();
  mqttPublishSendorBME(bmeSensor1, mqttSensor1, stateSensor1, ts_APIKey_Sensor1, false);
  mqttPublishSendorDHT(dhtSensor2, mqttSensor2, ts_APIKey_Sensor2, true, (ts_APIKey_Sensor2 == mqttHeaterSensor));
  mqttPublishSendorDHT(dhtSensor3, mqttSensor3, ts_APIKey_Sensor3, false, (ts_APIKey_Sensor3 == mqttHeaterSensor));
  mqttPublishSendorDHT(dhtSensor4, mqttSensor4, ts_APIKey_Sensor4, false, (ts_APIKey_Sensor4 == mqttHeaterSensor));

  Serial.println("Data publishing complete");
  Serial.println();
};

// =============================================================================================================================
// == Управление периферийными устройствами ====================================================================================
// =============================================================================================================================
void ledOn() {
  ledState = true;
  
  digitalWrite(pinLEDINT, LOW);
  digitalWrite(pinLED, HIGH);
}

void ledOff() {
  ledState = false;
  
  digitalWrite(pinLEDINT, HIGH);
  digitalWrite(pinLED, LOW);
}

void ledOnNoState() {
  digitalWrite(pinLEDINT, LOW);
  digitalWrite(pinLED, HIGH);
}

void ledRestoreState() {
  if (ledState) {
    digitalWrite(pinLEDINT, LOW);
    digitalWrite(pinLED, HIGH);
  }
  else {
    digitalWrite(pinLEDINT, HIGH);
    digitalWrite(pinLED, LOW);
  }
}

void ledToggle() {
  ledState = not ledState;

  if (ledState) {
    digitalWrite(pinLEDINT, LOW);
    digitalWrite(pinLED, HIGH);
  }
  else {
    digitalWrite(pinLEDINT, HIGH);
    digitalWrite(pinLED, LOW);
  }
}

// =============================================================================================================================
// == Управление нагревателем ==================================================================================================
// =============================================================================================================================
void relayOn() {
  digitalWrite(pinRELAY, HIGH);
}

void relayOff() {
  digitalWrite(pinRELAY, LOW);
}

void heaterToggle(const bool newState, const bool publishData) {
  // Переключаем реле
  heater_State = newState;
  
  if (heater_State == true) {
    relayOn();
    Serial.println("Set value 'heater_State' = true :: Нагрев включен!");
  }
  else {
    relayOff();
    Serial.println("Set value 'heater_State' = false :: Нагрев отключен!");
  };

  // Публикуем новые данные
  if (publishData) {
    // Публикуем состояние на MQTT-брокере
    mqttPublishHeaterState();

    // Публикуем состояние на ThingSpeak
    if (publThingSpeak) {
      // Формируем GET запрос
      String payload = String(ts_ApiKey) + mqttHeaterSensor;
      if (heater_State) {
        payload = payload + ts_ApiField + "5=1";
      }
      else {
        payload = payload + ts_ApiField + "5=0";     
      };
      tsPublishData(payload.c_str());
    };

    Serial.println();
  };
}

// =============================================================================================================================
// == MQTT: подписка на команды ================================================================================================
// =============================================================================================================================

void mqttSubscribe(const char* topic) {
  mqttClient.subscribe(topic);
  Serial.print("MQTT :: subscribe to topic '");
  Serial.print(topic);
  Serial.println("'");
}

void mqttSubscribeTopic(const char* topic1, const char* topic2) {
  String topicSub = "";
  if (topic2 == "") {
    topicSub = mqttGetDeviceTopic(topic1);
  }
  else {
    topicSub = mqttGetDeviceTopic(topic1, topic2);
  };
 
  mqttSubscribe(topicSub.c_str());
}

// =============================================================================================================================
// == MQTT: реакция на внешние команды =========================================================================================
// =============================================================================================================================
// void mqttCallback(char* topic, byte* payload, unsigned int length) {
void mqttCallback(const MQTT::Publish& pub) {
  ledOnNoState();
  
  Serial.print("MQTT :: message arrived [");
  Serial.print(pub.topic());
  Serial.print("] ");
  Serial.println(pub.payload_string());

  // == Синхронизация устройства ===============================================================================================
  
  // Запрос статуса устройства
  if (syncRequests && (pub.topic().equalsIgnoreCase(mqttWill))) {
    if (pub.payload_string().equalsIgnoreCase(mqttSyncRequest)) {
      mqttPublishDeviceState();
    }
    else {
      Serial.println("Ignored");
      Serial.println();
    }
  };

  // Режим синхронизации
  if (pub.topic().equalsIgnoreCase(mqttGetDeviceTopic(mqttSettings, mqttSyncMode))) {
    syncRequests = (pub.payload_string().equalsIgnoreCase("1") || pub.payload_string().equalsIgnoreCase("Включено"));
    EEPROM_bool_write(syncRequests_Addr, syncRequests);
    EEPROM.commit();
    Serial.print("Set value 'syncRequests' = ");
    Serial.println(syncRequests);
    Serial.println("");
  };

  // == "Обычные" параметры устройства =========================================================================================

  // Интервал публикации данных
  if (pub.topic().equalsIgnoreCase(mqttGetDeviceTopic(mqttSettings, mqttPubInterval))) {
    intvPublish = pub.payload_string().toInt();
    EEPROM_uint16_write(intvPublish_Addr, intvPublish);
    EEPROM.commit();
    Serial.print("Set value 'intvPublish' = ");
    Serial.println(intvPublish);
    Serial.println("");
  };

  // Режим ThingSpeak
  if (pub.topic().equalsIgnoreCase(mqttGetDeviceTopic(mqttSettings, mqttThingSpeak))) {
    publThingSpeak = (pub.payload_string().equalsIgnoreCase("1") || pub.payload_string().equalsIgnoreCase("Включено"));
    EEPROM_bool_write(publThingSpeak_Addr, publThingSpeak);
    EEPROM.commit();
    Serial.print("Set value 'publThingSpeak' = ");
    Serial.println(publThingSpeak);
    Serial.println("");
  };

  // Переключение режима нагрева
  if (pub.topic().equalsIgnoreCase(mqttGetDeviceTopic(mqttHeater, mqttHeaterMode))) {
    if (pub.payload_string().equalsIgnoreCase("0") || pub.payload_string().equalsIgnoreCase("Отключено")) {
      heater_Mode = 0;
    };
    if (pub.payload_string().equalsIgnoreCase("1") || pub.payload_string().equalsIgnoreCase("Включено")) {
      heater_Mode = 1;
    };
    if (pub.payload_string().equalsIgnoreCase("2") || pub.payload_string().equalsIgnoreCase("Автоматически")) {
      heater_Mode = 2;
    };
    
    EEPROM_uint8_write(heater_Mode_Addr, heater_Mode);
    EEPROM.commit();
    Serial.print("Set value 'heater_Mode' = ");
    Serial.println(heater_Mode);

    heaterToggle(heater_Mode == 1, true);  
  };

  // Настройка границ температурного режима
  if (pub.topic().equalsIgnoreCase(mqttGetDeviceTopic(mqttHeater, mqttHeaterMin))) {
    heater_AutoOn = pub.payload_string().toInt();
    EEPROM_uint8_write(heater_AutoOn_Addr, heater_AutoOn);
    EEPROM.commit();
    Serial.print("Set value 'heater_AutoOn' = ");
    Serial.println(heater_AutoOn);
    Serial.println("");
  };
  
  if (pub.topic().equalsIgnoreCase(mqttGetDeviceTopic(mqttHeater, mqttHeaterMax))) {
    heater_AutoOff = pub.payload_string().toInt();
    EEPROM_uint8_write(heater_AutoOff_Addr, heater_AutoOff);
    EEPROM.commit();
    Serial.print("Set value 'heater_AutoOff' = ");
    Serial.println(heater_AutoOff);
    Serial.println("");
  };

  // == Команды ================================================================================================================
  if (pub.topic().equalsIgnoreCase(mqttGetDeviceTopic(mqttSettings, mqttCommands))) {
    if (pub.payload_string().equalsIgnoreCase(mqttCmdReset)) {
      Serial.println("Reset ESP!!!");
      Serial.println("");
      ESP.reset();
    } 
    else {
      Serial.print("Команда '");
      Serial.print(pub.payload_string());
      Serial.println("' не обработана!");
      Serial.println();
    };
  };

  ledRestoreState();
} 

// =============================================================================================================================
// == Инициализация устройства =================================================================================================
// =============================================================================================================================
void setup() {
  Serial.begin(9600);
  Serial.println("Initialization...");

  // Настройка портов
  Serial.println("Inititailize PINs...");
  pinMode(pinLED, OUTPUT);
  pinMode(pinLEDINT, OUTPUT);
  pinMode(pinRELAY, OUTPUT);
  ledState = false;

  // Считываем пераметры из памяти
  Serial.println("Read settings from EEPROM...");
  EEPROM.begin(512); 

  Serial.print("heater_Mode = ");
  heater_Mode = EEPROM_uint8_read(heater_Mode_Addr, heater_Mode);
  Serial.println(heater_Mode);

  Serial.print("heater_AutoOn = ");
  heater_AutoOn = EEPROM_uint8_read(heater_AutoOn_Addr, heater_AutoOn);
  Serial.println(heater_AutoOn);

  Serial.print("heater_AutoOff = ");
  heater_AutoOff = EEPROM_uint8_read(heater_AutoOff_Addr, heater_AutoOff);
  Serial.println(heater_AutoOff);

  Serial.print("intvPublish = ");
  intvPublish = EEPROM_uint16_read(intvPublish_Addr, intvPublish);
  Serial.println(intvPublish);

  Serial.print("publThingSpeak = ");
  publThingSpeak = EEPROM_bool_read(publThingSpeak_Addr, publThingSpeak);
  Serial.println(publThingSpeak);

  Serial.print("syncRequests = ");
  syncRequests = EEPROM_bool_read(syncRequests_Addr, syncRequests);
  Serial.println(syncRequests);

  EEPROM.commit();

  // Настраиваем WiFi подключение
  Serial.println("Inititailize WiFi connection...");
  WiFi.mode(WIFI_STA);
  wifiMulti.addAP(wifi1_ssid, wifi1_pass);
  wifiMulti.addAP(wifi2_ssid, wifi2_pass);
  wifiMulti.addAP(wifi3_ssid, wifi3_pass);
  wifiState = false;

  // Настраиваем MQTT подключение
  Serial.println("Inititailize MQTT connection...");
  Serial.print("ClientId: ");
  Serial.println(mqttDeviceId);
  mqttClient.set_server(mqttServer, mqttPort);
  mqttWill = String(mqttPrefix) + "/" + mqttDeviceId + "/" + mqttDeviceState;

  // Настраиваем I2C
  Serial.println("Inititailize sensors...");
  Wire.begin();
  Wire.setClock(100000); 

  // Подключаем уличный датчик BME280
  stateSensor1 = bmeSensor1.begin(0x76);
  if (stateSensor1) {
    Serial.println("Outdoor BME280 sensor connected");
  }
  else {
    Serial.println("Could not find a valid BME280 sensor, check wiring!");
  };

  // Подключаем датчики температуры DHT
  dhtSensor2.setup(pinSENS2, DHTesp::DHT22);
  dhtSensor3.setup(pinSENS3, DHTesp::AM2302);
  dhtSensor4.setup(pinSENS4, DHTesp::AM2302);
  
  // Инициализируем нагреватель
  Serial.println("Inititailize heater...");
  for (int i = 0; i < 3; i++) {
    relayOn();
    ledOff();
    delay(1000);
    
    relayOff();
    ledOn();
    delay(1000);
  };
  
  if ((heater_Mode == 0) || (heater_State == 2)) {
    heaterToggle(false, false);
  }
  else {
    heaterToggle(true, false);
  }
  
  lastPublish = millis();
  
  Serial.println("Initialization complete");
}

// =============================================================================================================================
// == Главный цикл =============================================================================================================
// =============================================================================================================================
void loop() {
  // Проверяем статус подключения к WiFi
  if (wifiMulti.run() == WL_CONNECTED) {
    // Если подключение только что выполнено, выводим сообщение
    if (!wifiState) {
      wifiState = true;

      Serial.println("");
      Serial.print("WiFi connected, local IP address: ");
      Serial.println(WiFi.localIP());     

      ledOff();
    }

    // Восстанавливаем подключение к MQTT (при необходимости)
    if (!mqttClient.connected()) {
      while (!mqttClient.connected()) {
        // Непрерывно мигаем светодиодом, пока нет подключения
        ledToggle();

        // Пробуем подключиться к MQTT брокеру
        Serial.print("Connecting to MQTT server '"); 
        Serial.print(mqttServer);
        Serial.print(":");
        Serial.print(mqttPort);
        Serial.print("'... ");
        
        // if (mqttClient.connect(mqtt_clientid, mqtt_user, mqtt_pass, mqttWill.c_str(), MQTTQOS2, true, "0")) {
        MQTT::Connect mqttCon = MQTT::Connect(mqttDeviceId);
        mqttCon.set_auth(mqttUser, mqttPass);
        mqttCon.set_will(mqttWill.c_str(), "0");
        
        if (mqttClient.connect(mqttCon)) {
          Serial.println("complete");
          Serial.println();

          // Публикуем текущий статус
          mqttPublishDeviceState();
          mqttPublishHeaterState();
          Serial.println("");

          // Подписываемся на внешние команды
          mqttClient.set_callback(mqttCallback);
          mqttSubscribe(mqttWill.c_str());
          mqttSubscribeTopic(mqttSettings, mqttSyncMode);
          mqttSubscribeTopic(mqttSettings, mqttPubInterval);
          mqttSubscribeTopic(mqttSettings, mqttThingSpeak);
          mqttSubscribeTopic(mqttHeater, mqttHeaterMode);
          mqttSubscribeTopic(mqttHeater, mqttHeaterMin);
          mqttSubscribeTopic(mqttHeater, mqttHeaterMax);
          mqttSubscribeTopic(mqttSettings, mqttCommands);
          
          Serial.println("");
        }
        else {
          Serial.print("failed, try again in 1 second"); 
          delay(1000);          
        }
      };

      ledOff();
    };

    // Обрабатываем MQTT подписки
    mqttClient.loop();

    // Публикуем новые данные
    if (millis() - lastPublish >= intvPublish * 1000) {
      ledOn();

      lastPublish = millis();
      mqttPublishSensorsData();
      
      ledOff(); 
    }
  }
  else {
    // Выводим надпись, что подключение к WiFi потеряно
    if (wifiState) {
      wifiState = false;
      Serial.println("WiFi connection lost!");
    };

    // Непрерывно мигаем светодиодом, пока нет подключения
    ledToggle();
    delay(250);
  } 
  yield();
}
